"""
basic functions: format, validate, display tracking
"""

import os
import re
from datetime import datetime
from pathlib import Path

from PySide2 import QtCore
from typing import Tuple, Union

from PySide2.QtWidgets import QTextBrowser
from obspy import UTCDateTime
import numpy as np
from sohstationviewer.view.util.enums import LogType


def validate_file(path2file: Union[str, Path], file_name: str):
    """
    Check if fileName given is a file and not info file
    :param path2file: absolute path to file
    :param file_name: name of the file
    :return: True if pass checking, False if not.
    """

    if file_name.strip() == '.DS_Store' or file_name.startswith('._'):
        # skip mac's info file
        return False

    if not os.path.isfile(path2file):
        return False
    return True


@QtCore.Slot()
def display_tracking_info(tracking_box: QTextBrowser, text: str,
                          type: LogType = LogType.INFO):
    """
    Display text in the given widget with different background and text colors
    :param tracking_box: widget to display tracking info
    :param text: info to be displayed
    :param type: (info/warning/error) type of info to be displayed in
        different color
    """
    if tracking_box is None:
        print(f"{type.name}: {text}")
        return

    msg = {'text': text}
    if type == LogType.ERROR:
        msg['color'] = 'white'
        msg['bgcolor'] = '#e46269'
    elif type == LogType.WARNING:
        msg['color'] = '#ffd966'
        msg['bgcolor'] = 'orange'
    else:
        msg['color'] = 'blue'
        msg['bgcolor'] = 'white'
    html_text = """<body>
        <div style='color:%(color)s; background-color:%(bgcolor)s'>
            %(text)s
        </div>
        </body>"""
    tracking_box.setHtml(html_text % msg)
    # parent.update()
    tracking_box.repaint()


def get_dir_size(dir: str) -> Tuple[int, int]:
    """
    Get size of directory and size of file.
    :param dir: absolute path to directory
    :return:
        total_size: total size of the directory
        total_file: total file of the directory
    """

    total_size = 0
    total_file = 0
    for path, subdirs, files in os.walk(dir):
        for file_name in files:
            if not validate_file(os.path.join(path, file_name), file_name):
                continue
            fp = os.path.join(path, file_name)
            total_size += os.path.getsize(fp)
            total_file += 1
    return total_size, total_file


def get_time_6(time_str: str) -> Tuple[float, int]:
    """
    Get time from 6 parts string.
    (year:day of year:hour:minute:second:millisecond)
    Ex: 01:251:09:41:35:656/ 2001:251:09:41:35:656
    in which year in the first part can be 2 digits or 6 digits
    :param time_str: 6 part time string
    :return the epoch time and the year of time_str.
    """
    year = time_str.split(':')[0]
    if len(year) == 2:
        return get_time_6_2y(time_str)
    else:
        return get_time_6_4y(time_str)


def get_time_6_2y(time_str: str) -> Tuple[float, int]:
    """
    Get time from 6 parts string in which year has 2 digits.
    Ex: 01:251:09:41:35:656
    :param time_str: 6 part time string with 2 digits for year
    :return the epoch time and the year of time_str.
    """
    # pad 0 so the last part has 6 digits to match with the format str
    time_str = time_str.ljust(22, "0")
    time = datetime.strptime(time_str, "%y:%j:%H:%M:%S:%f")
    utc_time = UTCDateTime(time)
    return utc_time.timestamp, time.year


def get_time_6_4y(time_str: str) -> Tuple[float, int]:
    """
    Get time from 6 parts string in which year has 4 digits.
    Ex: 2001:251:09:41:35:656
    :param time_str: 6 part time string with 4 digits for year
    :return the epoch time and the year of time_str.
    """
    # pad 0 so the last part has 6 digits to match with the format str
    time_str = time_str.ljust(24, "0")
    time = datetime.strptime(time_str, "%Y:%j:%H:%M:%S:%f")
    utc_time = UTCDateTime(time)
    return utc_time.timestamp, time.year


def get_time_4(time_str: str, tracking_year: int, y_added: bool
               ) -> Tuple[float, int, bool]:
    """
    Get time from 4 parts string. (day of year:hour:minute:second)
    Ex: 253:19:41:42
    :param time_str: time string
    :param tracking_year: year that has been detected
    :param y_added: flag to tell if year has been plussed 1 or not
    :return:
        + utc_time.timestamp: epoch time
        + time.year: year
        + y_added: flag to tell if year has been plussed 1 or not
    """
    if not y_added:
        # first day => move to next year
        doy = int(time_str.split(':')[0])
        if doy == 1:
            tracking_year += 1
            y_added = True
    time_str = f'{str(tracking_year)}:{time_str}'
    time = datetime.strptime(time_str, "%Y:%j:%H:%M:%S")
    utc_time = UTCDateTime(time)
    return utc_time.timestamp, time.year, y_added


def get_val(text: str) -> float:
    """
    Get the value part of a string with non-number substring following.
    :param text: value string including unit
    :return: value part including +/-, remove str that follows
    """
    re_val = '^\+?\-?[0-9]+\.?[0-9]?'            # noqa: W605
    return float(re.search(re_val, text).group())


def rtn_pattern(text: str, upper: bool = False) -> str:
    """
    This function is from logpeek's rtn_pattern.
    return routine pattern of the string with:
     + 0 for digit
     + a for lowercase
     + A for upper case
     + remain special character
    :param text: text to get format
    :param upper: flag of whether to convert all alphabetic characters
        to A
    :return rtn: routine pattern of the string
    """
    rtn = ""
    for c in text:
        if c.isdigit():
            rtn += "0"
        elif c.isupper():
            rtn += "A"
        elif c.islower():
            rtn += "a"
        else:
            rtn += c
    # So the A/a chars will always be A, so the caller knows what to look for
    if upper is True:
        return rtn.upper()
    return rtn


def add_thousand_separator(value: float) -> str:
    """
    This function is from logpeek's fmti
    Given Value will be convert to a string integer with thousand separators
    :param value: string of value with unit
    :return new_value: new value with no unit and with thousand separators
    """
    value = int(value)
    if value > -1000 and value < 1000:
        return str(value)
    value = str(value)
    new_value = ""
    # There'll never be a + sign.
    if value[0] == "-":
        offset = 1
    else:
        offset = 0
    count_digits = 0
    for i in range(len(value) - 1, -1 + offset, -1):
        new_value = value[i] + new_value
        count_digits += 1
        if count_digits == 3 and i != 0:
            new_value = "," + new_value
            count_digits = 0
    if offset != 0:
        if new_value.startswith(","):
            new_value = new_value[1:]
        new_value = value[0] + new_value
    return new_value


def apply_convert_factor(c_data: dict, convert_factor: float):
    """
    convertFactor = 150mV/count = 150V/1000count
    => unit data * convertFactor= data *150/1000 V

    :param c_data: data of the channel which includes down-sampled
        data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
        soh_data[key][chan_id]
    :param convert_factor: convertFactor field retrieved from
        db table Channels for this channel
    """
    c_data['data'] = np.multiply(c_data['data'], [convert_factor])
