#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Suggested updates to obspy.io.reftek.core:
- modify section of the code that deals with upacking data
  with '16' and '32' encodings to account for number of samples in each data
  packet. Mass position channels data are '16' encoded.

Maeva Pourpoint IRIS/PASSCAL
"""


import copy
import obspy.io.reftek.core as obspy_rt130_core
import warnings

import numpy as np

from obspy import Trace, Stream, UTCDateTime
from obspy.core.util.obspy_types import ObsPyException
from obspy.io.reftek.packet import _unpack_C0_C2_data
from sohstationviewer.model.reftek.from_rt2ms.packet import EHPacket


class Reftek130Exception(ObsPyException):
    pass


class Reftek130(obspy_rt130_core.Reftek130):

    def to_stream(self, network="", location="", component_codes=None,
                  include_mp123=False, include_mp456=False,
                  headonly=False, verbose=False,
                  sort_permuted_package_sequence=False):
        """
        :type headonly: bool
        :param headonly: Determines whether or not to unpack the data or just
            read the headers.
        """
        if verbose:
            print(self)
        if not len(self._data):
            msg = "No packet data in Reftek130 object (file: {})"
            raise Reftek130Exception(msg.format(self._filename))
        self.check_packet_sequence_and_sort(sort_permuted_package_sequence)
        self.check_packet_sequence_contiguous()
        self.drop_not_implemented_packet_types()
        if not len(self._data):
            msg = ("No packet data left in Reftek130 object after dropping "
                   "non-implemented packets (file: {})").format(self._filename)
            raise Reftek130Exception(msg)
        st = Stream()
        for event_number in np.unique(self._data['event_number']):
            data = self._data[self._data['event_number'] == event_number]
            # we should have exactly one EH and one ET packet, truncated data
            # sometimes misses the header or trailer packet.
            eh_packets = data[data['packet_type'] == b"EH"]
            et_packets = data[data['packet_type'] == b"ET"]
            if len(eh_packets) == 0 and len(et_packets) == 0:
                msg = ("Reftek data (file: {}) contain data packets without "
                       "corresponding header or trailer packet."
                       .format(self._filename))
                raise Reftek130Exception(msg)
            if len(eh_packets) > 1 or len(et_packets) > 1:
                msg = ("Reftek data (file: {}) contain data packets with "
                       "multiple corresponding header or trailer packets."
                       .format(self._filename))
                raise Reftek130Exception(msg)
            if len(eh_packets) != 1:
                msg = ("No event header (EH) packets in packet sequence. "
                       "File ({}) might be truncated.".format(self._filename))
                warnings.warn(msg)
            if len(et_packets) != 1:
                msg = ("No event trailer (ET) packets in packet sequence. "
                       "File ({}) might be truncated.".format(self._filename))
                warnings.warn(msg)
            # use either the EH or ET packet, they have the same content (only
            # trigger stop time is not in EH)
            if len(eh_packets):
                eh = EHPacket(eh_packets[0])
            else:
                eh = EHPacket(et_packets[0])
            # only C0, C2, 16, 32 encodings supported right now
            if eh.data_format == b"C0":
                encoding = 'C0'
            elif eh.data_format == b"C2":
                encoding = 'C2'
            elif eh.data_format == b"16":
                encoding = '16'
            elif eh.data_format == b"32":
                encoding = '32'
            else:
                msg = ("Reftek data encoding '{}' not implemented yet. Please "
                       "open an issue on GitHub and provide a small (< 50kb) "
                       "test file.").format(eh.data_format)
                raise NotImplementedError(msg)
            header = {
                "unit_id": self._data['unit_id'][0],
                "experiment_number": self._data['experiment_number'][0],
                "network": network,
                "station": (eh.station_name +
                            eh.station_name_extension).strip(),
                "location": location, "sampling_rate": eh.sampling_rate,
                "reftek130": eh._to_dict()}
            delta = 1.0 / eh.sampling_rate
            delta_nanoseconds = int(delta * 1e9)
            inds_dt = data['packet_type'] == b"DT"
            data_channels = np.unique(data[inds_dt]['channel_number'])
            for channel_number in data_channels:
                inds = data['channel_number'] == channel_number
                # channel number of EH/ET packets also equals zero (one of the
                # three unused bytes in the extended header of EH/ET packets)
                inds &= data['packet_type'] == b"DT"
                packets = data[inds]

                # split into contiguous blocks, i.e. find gaps. packet sequence
                # was sorted already..
                endtimes = (
                    packets[:-1]["time"] +
                    packets[:-1]["number_of_samples"].astype(np.int64) *
                    delta_nanoseconds)
                # check if next starttime matches seamless to last chunk
                # 1e-3 seconds == 1e6 nanoseconds is the smallest time
                # difference reftek130 format can represent, so anything larger
                # or equal means a gap/overlap.
                time_diffs_milliseconds_abs = np.abs(
                    packets[1:]["time"] - endtimes) / 1000000
                gaps = time_diffs_milliseconds_abs >= 1
                if np.any(gaps):
                    gap_split_indices = np.nonzero(gaps)[0] + 1
                    contiguous = np.array_split(packets, gap_split_indices)
                else:
                    contiguous = [packets]

                for packets_ in contiguous:
                    starttime = packets_[0]['time']

                    if headonly:
                        sample_data = np.array([], dtype=np.int32)
                        npts = packets_["number_of_samples"].sum()
                    else:
                        if encoding in ('C0', 'C2'):
                            sample_data = _unpack_C0_C2_data(packets_,
                                                             encoding)
                        elif encoding in ('16', '32'):
                            # rt130 stores in big endian
                            dtype = {'16': '>i2', '32': '>i4'}[encoding]
                            # just fix endianness and use correct dtype
                            sample_data = np.require(
                                packets_['payload'],
                                requirements=['C_CONTIGUOUS'])
                            # either int16 or int32
                            sample_data = sample_data.view(dtype)
                            # account for number of samples, i.e. some packets
                            # might not use the full payload size but have
                            # empty parts at the end that need to be cut away
                            number_of_samples_max = sample_data.shape[1]
                            sample_data = sample_data.flatten()
                            # go through packets starting at the back,
                            # otherwise indices of later packets would change
                            # while looping
                            for ind, num_samps in reversed([
                                    (ind, num_samps) for ind, num_samps in
                                    enumerate(packets_["number_of_samples"])
                                    if num_samps != number_of_samples_max]):
                                # looping backwards we can easily find the
                                # start of each packet, since the earlier
                                # packets are still untouched and at maximum
                                # sample length in our big array with all
                                # packets
                                start_of_packet = ind * number_of_samples_max
                                start_empty_part = start_of_packet + num_samps
                                end_empty_part = (start_of_packet +
                                                  number_of_samples_max)
                                sample_data = np.delete(
                                    sample_data,
                                    slice(start_empty_part, end_empty_part))
                        npts = len(sample_data)

                    tr = Trace(data=sample_data, header=copy.deepcopy(header))
                    # channel number is not included in the EH/ET packet
                    # payload, so add it to stats as well..
                    tr.stats.reftek130['channel_number'] = channel_number
                    if headonly:
                        tr.stats.npts = npts
                    tr.stats.starttime = UTCDateTime(ns=starttime)
                    """
                    if component codes were explicitly provided, use them
                    together with the stream label
                    if component_codes is not None:
                        tr.stats.channel = (eh.stream_name.strip() +
                                            component_codes[channel_number])
                    # otherwise check if channel code is set for the given
                    # channel (seems to be not the case usually)
                    elif eh.channel_code[channel_number] is not None:
                        tr.stats.channel = eh.channel_code[channel_number]
                    # otherwise fall back to using the stream label together
                    # with the number of the channel in the file (starting with
                    # 0, as Z-1-2 is common use for data streams not oriented
                    # against North)
                    else:
                        msg = ("No channel code specified in the data file "
                               "and no component codes specified. Using "
                               "stream label and number of channel in file as "
                               "channel codes.")
                        warnings.warn(msg)
                        tr.stats.channel = (
                            eh.stream_name.strip() + str(channel_number))
                    """
                    DS = self._data['data_stream_number'][0] + 1
                    if DS != 9:
                        tr.stats.channel = "DS%s-%s" % (DS, channel_number + 1)
                    else:
                        if not include_mp123 and channel_number < 3:
                            continue
                        if not include_mp456 and channel_number > 2:
                            continue
                        tr.stats.channel = "MP%s" % (channel_number + 1)
                        # check if endtime of trace is consistent
                    t_last = packets_[-1]['time']
                    npts_last = packets_[-1]['number_of_samples']
                    try:
                        if not headonly:
                            assert npts == len(sample_data)
                        if npts_last:
                            assert tr.stats.endtime == UTCDateTime(
                                ns=t_last) + (npts_last - 1) * delta
                        if npts:
                            assert tr.stats.endtime == (
                                tr.stats.starttime + (npts - 1) * delta)
                    except AssertionError:
                        msg = ("Reftek file has a trace with an inconsistent "
                               "endtime or number of samples. Please open an "
                               "issue on GitHub and provide your file for"
                               "testing.")
                        raise Reftek130Exception(msg)
                    st += tr
        return st
