"""
param_dialog.py
GUI to add/dit/remove params
NOTE: Cannot remove or change params that are already used for channels.
"""
from PySide2 import QtWidgets

from sohstationviewer.view.util.plot_func_names import plot_functions
from sohstationviewer.view.db_config.db_config_dialog import UiDBInfoDialog

from sohstationviewer.database.process_db import execute_db

from sohstationviewer.conf.dbSettings import dbConf


class ParamDialog(UiDBInfoDialog):
    def __init__(self, parent):
        """
        :param parent: QMainWindow/QWidget - the parent widget
        """
        self.require_valuecolors_plottypes = [
            p for p in plot_functions.keys()
            if 'ValueColors' in plot_functions[p][0]]
        super().__init__(
            parent,
            ['No.', 'Param', 'Plot Type', 'ValueColors', 'Height    '],
            'param', 'parameters',
            resize_content_columns=[0, 3])
        self.setWindowTitle("Edit/Add/Delete Parameters")

    def add_row(self, row_idx, fk=False):
        """
        Add a row of widgets to self.data_table_widgets.

        :param row_idx: index of row to add
        :param fk: bool: True if there is a foreign constrain that prevents the
            row to be deleted
        """
        self.add_widget(None, row_idx, 0)       # No.
        self.add_widget(self.data_list, row_idx, 1, foreign_key=fk)
        plot_type = self.add_widget(
            self.data_list, row_idx, 2,
            choices=[''] + sorted(plot_functions.keys()))
        place_holder_text = ""
        if plot_type in self.require_valuecolors_plottypes:
            place_holder_text = "Enter ValueColors..."
            if plot_type == 'linesDots':
                place_holder_text = "Ex: L:G|D:W"
            elif plot_type == 'multiColorDots':
                place_holder_text = "Ex: -1:_|0:R|2.3:Y|+2.3:G"
            elif plot_type == "upDownDots":
                place_holder_text = "Ex: 1:R|0:Y"
            elif plot_type == "dotForTime":
                place_holder_text = "Ex: G"
        self.add_widget(self.data_list, row_idx, 3,
                        place_holder_text=place_holder_text)
        self.add_widget(self.data_list, row_idx, 4, range_values=[0, 10])

    def get_data_list(self):
        """
        Get list of data to fill self.data_table_widgets' content
        """
        param_rows = execute_db(
            "SELECT param,"
            " IFNULL(plotType, '') AS plotType,"
            " IFNULL(valueColors, '') AS valueColors,"
            " IFNULL(height, 0) AS height FROM Parameters")
        return [[d[0], d[1], d[2], int(d[3])]
                for d in param_rows]

    def get_row_inputs(self, row_idx):
        """
        Get content from a row of widgets.
        The valueColors string's format is checked here before entered into DB

        :param row_idx: index of row
        """
        # check value_colors string
        plot_type = self.data_table_widget.cellWidget(
            row_idx, 2).currentText()
        value_colors_string = self.data_table_widget.cellWidget(
            row_idx, 3).text().strip()
        if value_colors_string == "" and plot_type == 'multiColorDots':
            raise Exception(f"Row {row_idx}: multiColorDots type requires "
                            f"some ValueColors value")
        if value_colors_string != "":
            if plot_type in self.require_valuecolors_plottypes:
                value_colors = value_colors_string.split("|")
                try:
                    for vc in value_colors:
                        vc = vc.strip()
                        if not dbConf[plot_type]['pattern'].match(vc):
                            msg = (f"Row {row_idx}: The ValueColors got is "
                                   f"'{vc}' which doesn't match the required "
                                   f"format.\n\n"
                                   f"{dbConf[plot_type]['instruction']}")
                            raise Exception(msg)
                except KeyError:
                    pass
            else:
                if plot_type == "":
                    msg = (f"Row {row_idx}: No ValueColors should be entered "
                           f"when no Plot Type selected.")
                else:
                    msg = (f"Row {row_idx}: No ValueColors is required for "
                           f"Plot Type {plot_type}.")
                QtWidgets.QMessageBox.information(
                    self, "Clear ValueColors", msg)
                self.data_table_widget.cellWidget(row_idx, 3).clear()

        return [
            self.data_table_widget.cellWidget(row_idx, 1).text().strip(),
            self.data_table_widget.cellWidget(row_idx, 2).currentText(),
            self.data_table_widget.cellWidget(row_idx, 3).text().strip(),
            int(self.data_table_widget.cellWidget(row_idx, 4).value())
        ]

    def update_data(self, row, widget_idx, list_idx):
        """
        Prepare insert, update queries then update data of a row from
            self.data_table_widgets' content.

        :param row: list - data of a row
        :param widget_idx: index of row in self.data_table_widgets
        :param list_idx: index of row in self.data_list
        """
        insert_sql = (f"INSERT INTO Parameters VALUES"
                      f"('{row[0]}', '{row[1]}', '{row[2]}', {row[3]})")
        update_sql = (f"UPDATE Parameters SET param='{row[0]}', "
                      f"plotType='{row[1]}', valueColors='{row[2]}',"
                      f"height={row[3]} "
                      f"WHERE param='%s'")
        return super().update_data(
            row, widget_idx, list_idx, insert_sql, update_sql)
