from copy import deepcopy
from typing import Union, Dict, List, Set, Tuple

import numpy as np
from PySide2.QtWidgets import QTextBrowser

from sohstationviewer.controller.plotting_data import format_time
from sohstationviewer.controller.processing import detect_data_type
from sohstationviewer.model.data_type_model import DataTypeModel
from sohstationviewer.model.mseed.mseed import MSeed
from sohstationviewer.model.reftek.reftek import RT130


def extract_data_set_info(tracking_box: QTextBrowser,
                          data_obj: Union[DataTypeModel, RT130, MSeed],
                          date_format: str
                          ) -> Dict[str, Union[str, List[str]]]:
    """
    Extract information about a data set. The extracted information includes:
        - Sources read
        - Start / end times
        - Station ID
        - DAS Serial Number
        - Network Code
        - Tag number
        - GPS information
        - Software version

    :param tracking_box: widget to display tracking info on
    :param data_obj: the data object to extract file information from
    :param date_format: the format by which to format file's start/end times
    :return: a dictionary containing the information of a file
    """

    data_set_info = {}
    sources_read = data_obj.dir
    data_set_info['Sources Read'] = sources_read.name

    data_type = detect_data_type(tracking_box, [data_obj.dir])

    start_time, end_time = extract_start_end_time(data_obj)
    time_format = 'HH:MM:SS'
    start_time_str = format_time(start_time, date_format, time_format)
    end_time_str = format_time(end_time, date_format, time_format)
    data_set_info['Start time'] = start_time_str
    data_set_info['End time'] = end_time_str

    if data_type == 'RT130':
        das_serials = list({key[0] for key in data_obj.keys})
        data_set_info['DAS Serial Numbers'] = das_serials
    else:
        stations = list(data_obj.stats)
        data_set_info['Station IDs'] = stations

    if data_type != 'RT130':
        if data_obj.nets_in_file:
            networks = max(data_obj.nets_in_file, key=len)
            data_set_info['Network Codes'] = networks

    if data_type == 'Q330':
        tag_numbers, software_versions = extract_log_info_q330(data_obj)
        data_set_info['Tag Numbers'] = sorted(list(tag_numbers))
        data_set_info['Software Versions'] = sorted(list(software_versions))
    elif data_type == 'RT130':
        cpu_versions, gps_versions = extract_log_info_rt130(data_obj)
        data_set_info['CPU Software Versions'] = sorted(list(cpu_versions))
        data_set_info['GPS Firmware Versions'] = sorted(list(gps_versions))
    elif data_type == 'Pegasus':
        firmware_versions = extract_log_info_pegasus(data_obj)
        data_set_info['Firmware Versions'] = list(firmware_versions)

    return data_set_info


def extract_start_end_time(data_obj: DataTypeModel):
    """
    Extract start and end time of a data set.
    :param data_obj: the data object to extract start/end time from
    :return: the start and end time obtained from data_obj
    """
    start_time = float('inf')
    end_time = float('-inf')

    soh_start_time, soh_end_time = extract_start_end_time_soh(data_obj)
    if soh_start_time < start_time:
        start_time = soh_start_time
    if soh_end_time > end_time:
        end_time = soh_end_time

    if data_obj.waveform_data:
        wf_start_time, wf_end_time = extract_start_end_time_waveform(data_obj)
        if wf_start_time < start_time:
            start_time = wf_start_time
        if wf_end_time > end_time:
            end_time = wf_end_time

    return start_time, end_time


def extract_start_end_time_soh(data_obj: DataTypeModel):
    """
    Extract start and end time of a data set from its SOH data. Count mass
    position data as SOH data.
    :param data_obj: the data object that contains the data of the data set
    :return: the SOH data start and end time
    """
    start_time = float('inf')
    end_time = float('-inf')

    merged_data = deepcopy(data_obj.soh_data)
    for key, value in data_obj.mass_pos_data.items():
        if key in merged_data:
            merged_data[key].update(value)
        else:
            merged_data[key] = deepcopy(value)

    for key_data in merged_data.values():
        for chan, chan_data in key_data.items():
            trace = chan_data['orgTrace']
            trace_start_time = np.amin(trace['times'])
            if trace_start_time < start_time:
                start_time = trace_start_time
            trace_end_time = np.amax(trace['times'])
            if trace_end_time > end_time:
                end_time = trace_end_time
    return start_time, end_time


def extract_start_end_time_waveform(data_obj: DataTypeModel):
    """
    Extract start and end time of a data set from its waveform data.
    :param data_obj: the data object that contains the data of the data set
    :return: the waveform data start and end time
    """
    start_time = float('inf')
    end_time = float('-inf')

    for key_data in data_obj.waveform_data.values():
        files_info = key_data['filesInfo']
        for chan_data in files_info.values():
            for file_data in chan_data:
                if file_data['startEpoch'] < start_time:
                    start_time = file_data['startEpoch']
                if file_data['endEpoch'] > end_time:
                    end_time = file_data['endEpoch']

    return start_time, end_time


def extract_log_info_q330(data_obj: MSeed) -> Tuple[Set[str], Set[str]]:
    """
    Extract tag numbers and software versions from LOG channel of Q330 data.

    :param data_obj: the object that contains the Q330 data
    :return: the set of found tag numbers and the set of found software
        versions
    """
    tag_numbers = set()
    software_versions = set()
    log_str = '\n'.join(data_obj.log_data[data_obj.selected_key]['LOG'])
    log_lines = [line for line in log_str.splitlines() if line != '']
    for line in log_lines:
        if 'Tag Number' in line:
            line = line.split(': ')
            tag_numbers.add(line[1])
        if 'System Software Version' in line:
            line = line.split(': ')
            software_versions.add(line[1])
    return tag_numbers, software_versions


def extract_log_info_rt130(data_obj: RT130) -> Tuple[Set[str], Set[str]]:
    """
    Extract the CPU and GPS versions from the SOH data of RT130 data.

    :param data_obj: the object that contains the RT130 data
    :return: the set of found CPU versions and the set of found GPS versions
    """
    cpu_versions = set()
    gps_versions = set()
    log_str = '\n'.join(data_obj.log_data[data_obj.selected_key]['SOH'])
    log_lines = [line for line in log_str.splitlines() if line != '']
    for line in log_lines:
        if 'CPU SOFTWARE' in line:
            cpu_versions.add(' '.join(line.split()[3:]))
        if 'GPS FIRMWARE VERSION' in line:
            gps_version = ' '.join(line.split()[4:])
            # A data set (2017149.92EB) stores its GPS firmware version
            # followed by a bunch of commas. We remove those commas so that the
            # GPS firmware version of this data set looks similar to that of
            # other data sets.
            gps_version = gps_version.strip(',')
            gps_versions.add(gps_version)
    return cpu_versions, gps_versions


def extract_log_info_pegasus(data_obj: MSeed) -> Set[str]:
    """
    Extract the firmware versions from the log data of Pegasus data.

    :param data_obj: the object that contains the Pegasus data
    :return: the set of found firmware versions
    """
    firmware_versions = set()
    log_str = '\n'.join(data_obj.log_data['TEXT'])
    log_lines = [line for line in log_str.splitlines() if line != '']
    for line in log_lines:
        if 'Firmware version' in line:
            firmware_versions.add(line.split()[6])
    return firmware_versions
